/*****************************************************************************/
/*
                                   Net.h
*/
/*****************************************************************************/

#ifndef NET_H_LOADED
#define NET_H_LOADED 1

#include "wasd.h"

#define DEFAULT_HTTP_PORT         80
#define DEFAULT_HTTPS_PORT       443
#define DEFAULT_HTTP_PROXY_PORT 8080
#define DEFAULT_CONNECT_PORT    4443

#ifdef IP_UCX
#define CLIENT_SOCKET_IN_USE(reqptr) (reqptr->ClientChannel != 0)
#endif

#ifdef IP_NETLIB
#define CLIENT_SOCKET_IN_USE(reqptr) (reqptr->ClientNetLibSocket != NULL)
#define NETLIB_ADDRESS_LIST_SIZE 16
#endif

/**************/
/* structures */
/**************/

struct ServiceStruct
{
   struct ServiceStruct  *NextPtr;

   boolean  ConnectService,
            LogFileOpen,
            LogFileError,
            ProxyService,
            ProxyFileCacheEnabled;

   int  ClientPort,
        ConnectCount,
        LogFileNameLength,
        ProxyChainHostNameLength,
        ProxyChainPort,
        RequestScheme,
        ServerHostNameLength,
        ServerPort;

   unsigned long  ClientIpAddress,
                  ProxyChainIpAddress,
                  ServerIpAddress;

   unsigned long  QuadBytesRawRx [2],
                  QuadBytesRawTx [2];

   char  *RequestSchemeNamePtr;

   char  ClientHostName [128],
         ClientIpAddressString [32],
         LogFileName [256],
         LogHostName [64],
         ProxyChainHostName [128],
         ProxyChainHostPort [128+16],
         ProxyChainIpAddressString [32],
         ProxyChainPortString [16],
         ServerHostName [128],
         ServerHostPort [128+16],
         ServerIpAddressString [32],
         ServerPortString [16],
         SeSoLaCert [128],
         SeSoLaCipherList [256],
         SeSoLaKey [128];

   struct AnIOsb  ServerIOsb;

   struct dsc$descriptor  LogFileNameDsc;
   
   struct FAB  LogFileFab;
   struct RAB  LogFileRab;

   /* this will be cast using (SSL_CTX*) in SESOLA.C */
   void  *SeSoLaCtx;

#ifdef IP_UCX

   unsigned short  ClientChannel,
                   ServerChannel;

   struct sockaddr_in  ServerSocketName,
                       ClientSocketName;

   struct hostent  *HostEntryPtr;

   struct {
      unsigned int  Length;
      char  *Address;
   } ServerSocketNameItem;

   int  ClientSocketNameLength;
   struct {
      unsigned int  Length;
      char  *Address;
      unsigned int  *LengthPtr;
   } ClientSocketNameItem;

#endif /* IP_UCX */

#ifdef IP_NETLIB

   unsigned long  ClientNetLibSocket,
                  ServerNetLibSocket;

   struct SINDEF  ServerNetLibName,
                  ClientNetLibName;

   struct INADDRDEF NetLibAddressList [16];

   unsigned long  NetLibReturnedLength;
   struct dsc$descriptor  ClientIpAddressStringDsc;
   struct dsc$descriptor  ServerHostNameDsc;
   struct dsc$descriptor  ServerIpAddressStringDsc;

#endif /* IP_NETLIB */

};

/***********************/
/* function prototypes */
/***********************/

#ifdef IP_UCX
   NetAccessDenied (unsigned short, char*, char*);
   NetDirectWrite (unsigned short, char*, int);
   NetDummyRead (unsigned short);
   NetDummyReadAst (unsigned short);
   NetTooBusy (unsigned short, char*, char*);
#endif

#ifdef IP_NETLIB
   NetAccessDenied (unsigned long*, char*, char*);
   NetDirectWrite (unsigned long*, char*, int);
   NetDummyRead (unsigned long*);
   NetDummyReadAst (unsigned long*);
   NetTooBusy (unsigned long*, char*, char*);
#endif

int NetWriteBuffered (struct RequestStruct*, void*, char*, int);
int NetWriteDefered (struct RequestStruct*, void*, char*, int);
int NetWriteDeferedBuffers (struct RequestStruct*);
NetAccept (struct ServiceStruct*);
NetAcceptAst (struct ServiceStruct*);
NetBreakSocket (struct RequestStruct*);
NetBreakSocketAst (struct RequestStruct*);
NetClientAddressLookup (struct ServiceStruct*);
NetCreateService ();
NetGetServerHostName ();
NetHostNameLookup (char*, char*, char*, char*, int*, char*, int*);
NetListen ();
NetServerShutdown (unsigned short);
NetResponseHeaderAst (struct RequestStruct*);
NetServiceConfigured (int, int);
NetSetService ();
NetShutdownSocket (struct RequestStruct*, void*);
NetShutdownSocketAst (struct RequestStruct*);
NetCloseSocket (struct RequestStruct*);
int NetRead (struct RequestStruct*, void*, char*, int);
NetReadRawAst (struct RequestStruct*);
int NetWrite (struct RequestStruct*, void*, char*, int);
NetWriteRawAst (struct RequestStruct*);

#endif /* NET_H_LOADED */

/*****************************************************************************/

