/*****************************************************************************/
/*
                                ProxyStruct.h

08-AUG-98  MGD  initial
*/
/*****************************************************************************/

#ifndef PROXYSTRUCT_H_LOADED
#define PROXYSTRUCT_H_LOADED 1

#include <atrdef.h>
#include <fibdef.h>

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

/********************/
/* proxy accounting */
/********************/

struct ProxyAccountingStruct
{
boolean  FreeSpaceAvailable,
         ServingEnabled;

int  CacheReadCount,
     CacheRead304Count,
     CacheWriteCount,
     LookupCacheCount,
     LookupDnsCount,
     LookupNoneCount,
     LookupErrorCount,
     MethodConnectCount,
     MethodDeleteCount,
     MethodGetCount,
     MethodHeadCount,
     MethodPostCount,
     MethodPutCount,
     NetworkCount,
     RequestNotCacheableCount,
     ResponseNotCacheableCount;

unsigned long  /* count of bytes received from proxied server */
               QuadBytesCacheRx [2],
               /* count of bytes sent to proxied server */
               QuadBytesCacheTx [2],
               /* bytes received from proxied server - not cacheable */
               QuadBytesNotCacheableRx [2],
               /* bytes sent to proxied server - not cacheable */
               QuadBytesNotCacheableTx [2],
               /* count of bytes in cache-serviced request header */
               QuadBytesRawRx [2],
               /* count of bytes in cache-serviced response */
               QuadBytesRawTx [2];

char  StatusString [255-(sizeof(int)*30)];
};

/**************************/
/* proxy cache structures */
/**************************/

/* i.e. originated and works with any after v6.0.0 */
#define PROXY_CACHE_FILE_VERSION 0x060000

struct ProxyCacheDescrStruct
{
   int  CacheVersion,
        UrlLength,
        HeaderLength;
};

struct ProxyCacheRecordAttributes
{
   unsigned char  OfNoInterest1 [4];
   unsigned short  AllocatedVbnHi;
   unsigned short  AllocatedVbnLo;
   unsigned short  EndOfFileVbnHi;
   unsigned short  EndOfFileVbnLo;
   unsigned short  FirstFreeByte;
   unsigned char   OfNoInterest2 [18];
};

struct ProxyCacheAscDates
{
   unsigned short  RevisionCount;
   unsigned char  OfNoInterest [33];
};

struct ProxyCacheFileAcpStruct
{
   unsigned short  Channel,
                   AcpFileNameLength;
   unsigned long  CdtBinTime [2],
                  EdtBinTime [2],
                  RdtBinTime [2];
   char AcpFileName [128];
   struct ProxyCacheAscDates  AscDates;
   struct ProxyCacheRecordAttributes  RecAttr;
   struct fibdef  FileFib;
   struct atrdef  FileAtr [6];
   struct dsc$descriptor  FileFibAcpDsc,
                          FileNameAcpDsc;
   struct AnIOsb  AcpIOsb;
};

/************************/
/* proxy task structure */
/************************/

struct ProxyTaskStruct
{
   boolean  AdminWatchItem,
            NotCacheable,
            ParseInUse,
            ProxyCacheBeyondHeader,
            ProxyCacheFileExisted,
            ProxyCacheSuitable,
            ResponsePragmaNoCache,
            ResponseSetCookie,
            RequestPragmaNoCache;

   int  ConnectPort,
        ConnectIpAddress,
        ContentCount,
        HttpMethod,
        ProxyCacheFileNameLength,
        ProxyCacheFileSizeInBytes,
        ProxyCacheLastModifiedHours,
        ProxyCacheReadBytes,
        ProxyLookupRetryCount,
        ProxyRequestNumber,
        RebuiltRequestLength,
        RequestHostIpAddress,
        RequestHostNamePortLength,
        RequestHttpMethod,
        RequestHttpVersion,
        RequestPort,
        ResponseBodyLength,
        ResponseBufferCount,
        ResponseBufferCacheCount,
        ResponseBufferNetCount,
        ResponseBufferRemaining,
        ResponseBufferSize,
        ResponseBytes,
        ResponseConsecutiveNewLineCount,
        ResponseContentLength,
        ResponseHeaderLength,
        ResponseHttpVersion,
        ResponseStatusCode;

   unsigned long  BytesRawRx,
                  BytesRawTx,
                  RequestHostIpAddress32bit;

   unsigned long  ProxyCacheFileCdt [2],
                  ResponseExpiresBinaryTime [2],
                  ResponseLastModifiedBinaryTime [2];

   char  *ConnectHostNamePortPtr,
         *ContentPtr,
         *RebuiltRequestPtr,
         *ResponseBufferPtr,
         *ResponseBufferCachePtr,
         *ResponseBufferCurrentPtr,
         *ResponseBufferNetPtr,
         *RequestHttpCookiePtr,
         *RequestHttpMethodNamePtr,
         *RequestPathInfoPtr,
         *RequestQueryStringPtr,
         *ResolveHostWatchPtr,
         *ResponseHeaderPtr;

   char  ExpandedFileName [256],
         ProxyCacheDirectoryName [64],
         ProxyCacheFileName [128],
         ProxyCacheMd5HexDigest [33],
         RequestHostName [128],
         RequestHostNamePort [128+16],
         RequestPortString [16],
         RequestSchemeName [16],
         ResponseExpires [32],
         ResponseHttpProtocol [16],
         ResponseLastModified [48],
         ResponseStatusCodeString [16],
         ResponseStatusDescription [64];

   /* pointer to service struct */
   struct ServiceStruct  *ServicePtr;

   /* pointer to associated request (if any) */
   struct RequestStruct  *RequestPtr;

   /* structure is used to get/set file characteristics */
   struct  ProxyCacheFileAcpStruct  FileAcpData;

   /* AST function pointers */
   void (*ProxyReadRawAstFunctionPtr)(struct ProxyTaskStruct*);
   void (*ProxyWriteRawAstFunctionPtr)(struct ProxyTaskStruct*);
   void (*ProxyShutdownAstFunctionPtr)(struct ProxyTaskStruct*);

   /* I/O status blocks */
   struct AnIOsb  ProxyConnectIOsb,
                  ProxyLookupIOsb,
                  ProxyReadIOsb,
                  ProxyShutdownIOsb,
                  ProxyWriteIOsb;

   /* RMS structures */
   struct FAB  LoadFab;
   struct FAB  ReadFab;
   struct RAB  FileRab;
   struct NAM  FileNam;
   struct XABDAT  FileXabDat;
   struct XABPRO  FileXabPro;

#ifdef IP_UCX

   unsigned short  ProxyChannel;

   struct sockaddr_in  ProxySocketName;

   struct hostent  *HostEntryPtr;

   struct {
      unsigned int  Length;
      char  *Address;
   } ProxySocketNameItem;

/*
   int  ProxySocketNameLength;
   struct {
      unsigned int  Length;
      char  *Address;
      unsigned int  *LengthPtr;
   } ProxySocketNameItem;
*/

   /* for the asynchronous resolution of host name to address */
   int  ResolveHostAddrBuff[16];
   struct dsc$descriptor ResolveHostAddressDsc;
   struct dsc$descriptor ResolveHostNameDsc;

#endif /* IP_UCX */

#ifdef IP_NETLIB

   unsigned long  ProxyNetLibSocket,
                  ProxyNetLibAddressListCount;
   struct INADDRDEF ProxyNetLibAddressList [NETLIB_ADDRESS_LIST_SIZE];
   struct SINDEF  ProxyNetLibName;
   struct dsc$descriptor ResolveHostNameDsc;

#endif /* IP_NETLIB */
};

#endif /* PROXYSTRUCT_H_LOADED */

/*****************************************************************************/

